
//----------------------------------------------------------------------
// Copyright 1997, Gary Wu, gwu@cco.caltech.edu.
// ALL RIGHTS RESERVED
//
// File Name:   SoShadedLineSet.C
// Description: SoShadedLineSet, an Inventor node that uses texture
//              mapping hardware to simulate rendering of illuminated
//              field lines.
//----------------------------------------------------------------------

static char rcsid[] = "$Id: SoShadedLineSet.C,v 1.5 1998/02/27 22:13:19 gwu Exp $";

// ---------------------------------------------------------------------
// $Id: SoShadedLineSet.C,v 1.5 1998/02/27 22:13:19 gwu Exp $
// $Source: /usr/multires/projects/SoShadedLineSet/SoShadedLineSet.C,v $
// ---------------------------------------------------------------------
// $Log: SoShadedLineSet.C,v $
// Revision 1.5  1998/02/27 22:13:19  gwu
// Recovered the "lost" sensor to detect changes in the lineset data.
//
// Revision 1.4  1998/02/23 22:05:02  gwu
// Now correctly takes the model matrix into account when computing the viewing direction (for Head light).
//
// Revision 1.3  1998/02/06 02:35:14  gwu
// Updated the comments.  Added the man page in SoShadedLineSet.txt.
//
// Revision 1.2  1998/02/06 01:30:35  gwu
// Updated comment fields.  Put in version information placeholders.
//
//
// Initial version: gwu - CS174c Project, Spring 1997
// Revisions: 
//          2-4-98: gwu - Added proper destructor
//                        Build display lists for state change
//                        Used glBindTextureEXT to speed up texture swaps
// ---------------------------------------------------------------------



#include <assert.h>
#include <stdlib.h>

#include <sys/time.h>


#include <iostream.h>


#include <GL/gl.h>

#include <Inventor/actions/SoAction.h>
#include <Inventor/actions/SoGLRenderAction.h>
#include <Inventor/elements/SoViewingMatrixElement.h>
#include <Inventor/elements/SoModelMatrixElement.h>
#include <Inventor/elements/SoLazyElement.h>

#include <Inventor/sensors/SoFieldSensor.h>

#include <Inventor/bundles/SoMaterialBundle.h>

#include "SoShadedLineSet.h"



SO_NODE_SOURCE( SoShadedLineSet );



const SoCoordinateElement* SoShadedLineSet::s_CoordElement = NULL;



// Initializes the class
void SoShadedLineSet::initClass( void )
{
  s_CoordElement = NULL;

  // Initialize the member fields
  SO_NODE_INIT_CLASS( SoShadedLineSet, SoShape, "Shape" );

  SO_ENABLE( SoGLRenderAction, SoCoordinateElement );
}





SoShadedLineSet::SoShadedLineSet( void )
  : m_Texture( 0.1, 0.3, 0.6, 0.5 ),
    m_TextureMode_List( 0 ), 
    m_TextureBinding( 0 ), 
    m_AlphaIndex( 0 )
{
  SO_NODE_CONSTRUCTOR( SoShadedLineSet );
  SO_NODE_ADD_FIELD( numVertices, (-1) );
  SO_NODE_ADD_FIELD( L, (0.0, 0.0, -1.0) );
  SO_NODE_ADD_FIELD( V, (0.0, 0.0, -1.0) );
  SO_NODE_ADD_FIELD( stream, (FALSE) );
  SO_NODE_ADD_FIELD( startIndex, (0) );
  SO_NODE_ADD_FIELD( vertexProperty, (0) );


  if( SO_NODE_IS_FIRST_INSTANCE() )
  {
    // Set up the line texture
  }


  m_UpdateCache = 0;
  m_TexCoords = NULL;
  m_NumTexCoords = 0;

  m_SegmentList[0] = NULL;
  m_SegmentList[1] = NULL;
  m_SegmentList[2] = NULL;

  m_VertexAlphaIndex = NULL;


  SoFieldSensor* sensor = new SoFieldSensor( fieldSensorCB, this );
  sensor->attach( &numVertices );

}




SoShadedLineSet::~SoShadedLineSet()
{
  int i = 0;
  if( m_TexCoords )
  {
    for( i = 0; i < m_NumTexCoords; ++i )
    {
      delete[] m_TexCoords[i];
    }
    delete[] m_TexCoords;
  }

  for( i = 0; i < 3; ++i )
  {
    if( m_SegmentList[i] )
      delete[] m_SegmentList[i];
  }

  if( m_VertexAlphaIndex )
    delete[] m_VertexAlphaIndex;
  
}




void SoShadedLineSet::fieldSensorCB( void* data, SoSensor* )
{
  SoShadedLineSet* set = (SoShadedLineSet*) data;
  assert( set );
  set->m_UpdateCache = 0;
}





void SoShadedLineSet::GLRender( SoGLRenderAction* action )
{

  static int count = 0;
  static int lastTime = 0;


  SoState* state = action->getState();

  
  if( !shouldGLRender( action ) )
    return;



  const SoCoordinateElement* el = 
    SoCoordinateElement::getInstance( state );
  assert( el );
  int num = el->getNum();
  s_CoordElement = el;


  if( num < 2 )
    return;

  int i = startIndex.getValue();
  int curVertex = 0;




     
  // ------- save current state --------------------------
    

  // set light model to BASE_COLOR
  int32_t lightModel = SoLazyElement::getLightModel( state );
  SoLazyElement::setLightModel( state, SoLazyElement::BASE_COLOR );
    
  // Make sure first material is sent if necessary. We'll use
  //   the SoMaterialBundle class because it makes things very easy.
  SoMaterialBundle mb(action);
  mb.sendFirst();

    
  

  glPushAttrib( GL_ALL_ATTRIB_BITS );
    
    
    
    
  // ----- set new state ------------------------
    
        
  glMatrixMode( GL_TEXTURE );
  glPushMatrix();
  glPixelStorei( GL_UNPACK_ALIGNMENT, 1 );


  if( m_TextureMode_List == 0 )
  {
    m_TextureMode_List = glGenLists( 1 );
    glGenTexturesEXT( 1, &m_TextureBinding );

    glBindTextureEXT( GL_TEXTURE_2D, m_TextureBinding );
    glTexImage2D( GL_TEXTURE_2D, 0, 4, 
		  m_Texture.TextureSize(), 
		  m_Texture.TextureSize(),
		  0, GL_RGBA, GL_UNSIGNED_BYTE, 
		  m_Texture.Texture() );
    

    glNewList( m_TextureMode_List, GL_COMPILE );
    {

      glBindTextureEXT( GL_TEXTURE_2D, m_TextureBinding );
      
      glTexParameterf( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP );
      glTexParameterf( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP );
      glTexParameterf( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR );
      glTexParameterf( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR );
      glTexEnvi( GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE );
      glEnable( GL_TEXTURE_2D );

      // glEnable( GL_LINE_SMOOTH );
      glEnable( GL_BLEND );
      glBlendFunc( GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA );    
      // glHint( GL_LINE_SMOOTH_HINT, GL_DONT_CARE );

    }
    glEndList();

  }
  
  glCallList( m_TextureMode_List );

  
    
    
  SbMatrix viewMatrix = SoViewingMatrixElement::get( state );
  SbMatrix modelMatrix = SoModelMatrixElement::get( state );

  
  SbVec3f view( 0., 0., 1. );
  viewMatrix.inverse().multDirMatrix( view, view );
  modelMatrix.inverse().multDirMatrix( view, view );
  view.normalize();
  static GLfloat matrix[4][4] = 
  {
    { 0., 0., 0., 0. },
    { 0., 0., 0., 0. },
    { 0., 0., 0., 0. },
    { 1., 1., 0., 2. }
  };
    
    
  // headlight
  SbVec3f view_inv = view;
  
  matrix[0][0] = view_inv[0];
  matrix[1][0] = view_inv[1];
  matrix[2][0] = view_inv[2];
    
  matrix[0][1] = view_inv[0];
  matrix[1][1] = view_inv[1];
  matrix[2][1] = view_inv[2];
    
  glLoadMatrixf( &matrix[0][0] );
    
    


  if( m_UpdateCache == 0 )
    UpdateCache( el );
  assert( m_TexCoords );
  


  float color[4];
  SbColor diffuse = SoLazyElement::getDiffuse( state, 0 );
  color[0] = diffuse[0];
  color[1] = diffuse[1];
  color[2] = diffuse[2];
  color[3] = 1.0;

  


  
  // determine the best list to draw with
  int dir = LIST_X;
  float dir_value = 0.;
  for( i = 0; i < 3; ++i )
  {
    if( fabs( dir_value ) < fabs( view_inv[i] ) )
    {
      dir_value = view_inv[i];
      dir = i;
    }
  }


  int i_start = 0;
  int i_end = num;
  int i_inc = 1;
  if( dir_value < 0. )
  {
    i_start = num - 1;
    i_end = -1;
    i_inc = -1;
  }

  m_AlphaIndex = (m_AlphaIndex + NUM_ALPHA_SEGMENTS - 1) % NUM_ALPHA_SEGMENTS;

  glColor4fv( color );
  glBegin( GL_LINES );
  for( i = i_start; i != i_end; i += i_inc )
  {
    curVertex = m_SegmentList[dir][i];
    if( 0 <= curVertex && curVertex < num - 1 )
    {
      if( stream.getValue() )
      {
	color[3] = ( ( m_VertexAlphaIndex[ curVertex ] + m_AlphaIndex ) %
			   NUM_ALPHA_SEGMENTS ) / 
	  (float) ( NUM_ALPHA_SEGMENTS - 1 );
	glColor4fv( color );
      }

      glTexCoord3fv( m_TexCoords[ curVertex ] );
      glVertex3fv( el->get3( curVertex ).getValue() );
      
      ++curVertex;
      glTexCoord3fv( m_TexCoords[ curVertex ] );
      glVertex3fv( el->get3( curVertex ).getValue() );

    }
  }
  glEnd();






  // ----- display frame rate statistics ---------------

  ++count;

  timeval t;
  gettimeofday( &t );
  if( t.tv_sec != lastTime )
  {
    cout << count << " fps\n";
    cout.flush();

    count = 0;
    lastTime = t.tv_sec;
  }  



  
  // ----- restore original state ---------------------

  glPopAttrib();

  SoLazyElement::setLightModel( state, lightModel );


  s_CoordElement = NULL;
}




void SoShadedLineSet::UpdateCache( const SoCoordinateElement* el )
{

  assert( el );
  int num = el->getNum();

  int i = 0;
  int j = 0;
  int curVertex = 0;
  int k = 0;
  
  // free and reallocate the texture/tangent vector cache
  if( m_TexCoords )
  {
    for( k = 0; k < m_NumTexCoords; ++k )
    {
      delete[] m_TexCoords[k];
    }
    delete[] m_TexCoords;
  }

  m_NumTexCoords = num;
  m_TexCoords = new float*[ num ];
  assert( m_TexCoords );
  for( k = 0; k < num; ++k )
  {
    m_TexCoords[k] = new float[3];
    assert( m_TexCoords[k] );
  }

  
  // free and reallocate the segment lists
  for( i = 0; i < 3; ++i )
  {
    if( m_SegmentList[i] )
      delete[] m_SegmentList[i];
    m_SegmentList[i] = new int[ num ];
    assert( m_SegmentList[i] );

    for( j = 0; j < num; ++j )
      m_SegmentList[i][j] = j;
  }

  // free and reallocate the vertex transparency index
  if( m_VertexAlphaIndex )
    delete[] m_VertexAlphaIndex;
  m_VertexAlphaIndex = new int[ num ];
  assert( m_VertexAlphaIndex );

  for( i = 0; i < num; ++i )
  {
    m_VertexAlphaIndex[i] = i % NUM_ALPHA_SEGMENTS;
  }



  // make end of polylines -1 so it doesn't draw to the next polyline
  curVertex = 0;
  for( i = 0; i < numVertices.getNum(); ++i )
  {
    curVertex += numVertices[ i ];
    
    m_SegmentList[ LIST_X ][ curVertex - 1 ] = -1;
    m_SegmentList[ LIST_Y ][ curVertex - 1 ] = -1;
    m_SegmentList[ LIST_Z ][ curVertex - 1 ] = -1;

  }




  curVertex = 0;
  i = startIndex.getValue();
  while( i < numVertices.getNum() && curVertex < num )
  {
    
    j = 0;

    while( (j < numVertices[ i ] || numVertices[ i ] == -1 )
	   && curVertex < num )
    {

      if( ( numVertices[ i ] >= 2 ) || 
	  ( numVertices[ i ] == -1 && num - curVertex >= 2 ) )
      {
	// this right now only looks at one segment;
	// should average across two segments
	int k = (j == 0) ? curVertex + 1 : curVertex;
	SbVec3f T( el->get3( k ) - el->get3( k - 1 ) );
	(void) T.normalize();
	m_TexCoords[curVertex][0] = T[0];
	m_TexCoords[curVertex][1] = T[1];
	m_TexCoords[curVertex][2] = T[2];
      }

      ++curVertex;
      ++j;
    }
    ++i;
  }


  cout << "updated " << num << " line segments\n";
  cout.flush();


  // now sort in the three directions
  qsort( m_SegmentList[ LIST_X ], num, sizeof( int ),
	 SoShadedLineSet::SortByX );
  qsort( m_SegmentList[ LIST_Y ], num, sizeof( int ),
	 SoShadedLineSet::SortByY );
  qsort( m_SegmentList[ LIST_Z ], num, sizeof( int ),
	 SoShadedLineSet::SortByZ );
  

  m_UpdateCache = 1;
    
}



int SoShadedLineSet::SortByX( const void* x0, const void* x1 )
{
  return 
    s_CoordElement->get3( *(int*)x0 )[0] - 
    s_CoordElement->get3( *(int*)x1 )[0];
}

int SoShadedLineSet::SortByY( const void* x0, const void* x1 )
{
  return 
    s_CoordElement->get3( *(int*)x0 )[1] - 
    s_CoordElement->get3( *(int*)x1 )[1];
}

int SoShadedLineSet::SortByZ( const void* x0, const void* x1 )
{
  return 
    s_CoordElement->get3( *(int*)x0 )[2] - 
    s_CoordElement->get3( *(int*)x1 )[2];
}






void SoShadedLineSet::generatePrimitives( SoAction* action )
{

  SoPrimitiveVertex pv0;
  SoPrimitiveVertex pv1;

  SoState* state = action->getState();

  const SoCoordinateElement* el = 
    SoCoordinateElement::getInstance( state );
  assert( el );
  int num = el->getNum();

  int i = startIndex.getValue();
  int j = 0;
  int curVertex = 0;
  
  while( i < numVertices.getNum() && curVertex < num )
  {
    j = 0;
    while( (j < numVertices[ i ] || numVertices[ i ] == -1 )
	   && curVertex < num )
    {
      if( j > 0 )
      {
	pv0.setPoint( el->get3( curVertex - 1 ) );
	pv0.setTextureCoords( SbVec4f( 0., 0., 0., 1. ) );
	pv0.setMaterialIndex( 0 );
	pv0.setNormal( L.getValue() );

	pv1.setPoint( el->get3( curVertex ) );
	pv1.setTextureCoords( SbVec4f( 0., 0., 0., 1. ) );
	pv1.setMaterialIndex( 0 );
	pv1.setNormal( L.getValue() );
	
	invokeLineSegmentCallbacks( action, &pv0, &pv1 );
      }

      ++curVertex;

      ++j;
    }
    
    ++i;
  }
  
}





void SoShadedLineSet::computeBBox( SoAction* action, 
				   SbBox3f& box, 
				   SbVec3f& center )
{
  computeCoordBBox( action, -1, box, center );
}






